#include "cachelab.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <stdbool.h>

typedef struct CacheLine {            // 结构体：高速缓存行
    bool valid;             // 有效位
    unsigned tag;           // 标记位
    unsigned lastVis;       // 上一次访问的时间
} CacheLine;
typedef struct CacheSet {            // 结构体：高速缓存组
    CacheLine *lines;       // 组中的所有行
    unsigned count;         // 组中存放有效数据的行数
} CacheSet;
typedef struct Cache {            // 结构体：高速缓存
    CacheSet *sets;         // 高速缓存的所有组
    unsigned S, s;          // 组数
    unsigned E;             // 行数
    unsigned B, b;          // 块大小
    unsigned time;          // 高速缓存的“时间”
} Cache;

Cache ca;
int hitc, missc, evic;
bool verbose;

CacheLine *findLine(CacheSet s, unsigned tag) {
    for (unsigned i = 0; i < s.count; ++i)
        if (s.lines[i].valid && s.lines[i].tag == tag) {
            s.lines[i].lastVis = ca.time;
            return &s.lines[i];
        }
    return NULL;
}
void setLine(CacheLine *l, unsigned tag) {
    l->valid = 1;
    l->tag = tag;
    l->lastVis = ca.time;
}
void newBlock(CacheSet *s, unsigned tag) {
    setLine(&s->lines[s->count], tag);
    ++s->count;
}
CacheLine *findLRU(CacheSet s) {
    unsigned id = 0;
    for (unsigned i = 1; i < ca.E; ++i)
        if (s.lines[i].lastVis < s.lines[id].lastVis)
            id = i;
    return &s.lines[id];
}
void access(unsigned addr) {
    ++ca.time;
    unsigned tag = addr >> (ca.b + ca.s), set = (addr >> ca.b) & (ca.S - 1);
    CacheLine *l = findLine(ca.sets[set], tag);
    if (l) {
        if (verbose) printf(" hit");
        ++hitc;
        return;
    }
    if (verbose) printf(" miss");
    ++missc;
    CacheSet *s = &ca.sets[set];
    if (s->count == ca.E) {
        if (verbose) printf(" eviction");
        ++evic;
        CacheLine *ev = findLRU(*s);
        setLine(ev, tag);

    } else newBlock(s, tag);
}

void readTrace(FILE *fp) {
    char opt;
    unsigned addr, size;
    while (fscanf(fp, " %c %x,%u", &opt, &addr, &size) > 0) {
        if (opt == 'I') continue;
        if (verbose) printf(" %c %x,%u", opt, addr, size);
        if (opt == 'M') access(addr), access(addr);
        else if (opt == 'L') access(addr);
        else if (opt == 'S') access(addr);
        if (verbose) printf("\n");
    }
    fclose(fp);
}

void initCache(int s, int E, int b) {
    ca.s = s, ca.E = E, ca.b = b;
    ca.S = 1 << s, ca.B = 1 << b;
    ca.sets = (CacheSet *)malloc(sizeof(CacheSet) * ca.S);
    for (unsigned i = 0; i < ca.S; ++i) {
        ca.sets[i].count = 0;
        ca.sets[i].lines = (CacheLine *)malloc(sizeof(CacheLine) * E);
        for (unsigned j = 0; j < E; ++j)
            ca.sets[i].lines[j].valid = 0;
    }
}

void printHelp(char *file) {
    printf("Usage: %s [-hv] -s <num> -E <num> -b <num> -t <file>\n", file);
    printf("Options:\n");
    printf("-h         Print this help message.\n");
    printf("-v         Optional verbose flag.\n");
    printf("-s <num>   Number of set index bits.\n");
    printf("-E <num>   Number of lines per set.\n");
    printf("-b <num>   Number of block offset bits.\n");
    printf("-t <file>  Trace file.\n");
    printf("\n");
    printf("Examples:\n");
    printf("linux>  %s -s 4 -E 1 -b 4 -t traces/yi.trace\n", file);
    printf("linux>  %s -v -s 8 -E 2 -b 4 -t traces/yi.trace\n", file);
}

int main(int argc, char **argv) {
    int s = -1, E = -1, b = -1;
    char opt;
    FILE *fp = NULL;
    while ((opt = getopt(argc, argv, "hvs:E:b:t:")) != -1) {
        switch (opt) {
            case 'h':
                printHelp(argv[0]);
                return 0;
            case 'v':
                verbose = 1;
                break;
            case 's':
                s = atoi(optarg);
                break;
            case 'E':
                E = atoi(optarg);
                break;
            case 'b':
                b = atoi(optarg);
                break;
            case 't':
                fp = fopen(optarg, "r");
                if (fp == NULL) {
                    printf("%s: No such file or directory\n", optarg);
                    return 0;
                }
                break;
            default:
                printHelp(argv[0]);
                return 0;
        }
    }
    if (!~s || !~E || !~b || !fp) {
        printf("Missing required command line argument\n");
        printHelp(argv[0]);
        return 0;
    }
    initCache(s, E, b);
    readTrace(fp);
    printSummary(hitc, missc, evic);
    return 0;
}
